/*
This is a simple tool that can be used to perform UNIX "cat" like operations
for vmchannel devices since Windows doesn't provide easy access to these
devices on the command line.
*/

#include <stdio.h>
#include <tchar.h>
#include <Windows.h>
#include <InitGuid.h>
#include <SetupAPI.h>
#include <stdio.h>

#pragma comment(lib, "Setupapi.lib")

/* GUID for VirtIO serial devices */
DEFINE_GUID(VIRTIO_SERIAL_GUID, 0x6fde7521, 0x1b65, 0x48ae, 0xb6, 0x28, 0x80, 0xbe, 0x62, 0x1, 0x60, 0x26);

/* control code for the ioctl command to get info from the serial port */
#define IOCTL_GET_INFORMATION CTL_CODE(FILE_DEVICE_UNKNOWN, 0x800, METHOD_BUFFERED, FILE_ANY_ACCESS)

/* struct for the port info that comes from the serial device */
typedef struct
{
    UINT                id;
    BOOLEAN             out_vq_full;
    BOOLEAN             host_connected;
    BOOLEAN             guest_connected;
    CHAR                name[1];
} VIRTIO_PORT_INFO;

BOOL virtioserial_find(const char* name,
                       wchar_t* device_path,
                       DWORD device_path_size)
{
    BOOL success = FALSE; /* default returning an error */
    DWORD index = 0;
    HDEVINFO dev_info = NULL;

    /* Lookup the device GUID */
    dev_info = SetupDiGetClassDevs(&VIRTIO_SERIAL_GUID,
                                   NULL,
                                   NULL,
                                   DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);
    if(dev_info == INVALID_HANDLE_VALUE)
    {
        DWORD error =  GetLastError();
        fprintf(stderr, "SetupDiGetClassDevs Error %d", error);
        return FALSE;
    }

    /* Loop through all available VirtIO serial devices. */
    while(TRUE)
    {
        BOOL status = FALSE;
        DWORD size = 0;
        DWORD required_size = 0;
        SP_DEVICE_INTERFACE_DETAIL_DATA* detail_data = NULL;
        wchar_t current_device_path[MAX_PATH];
        HANDLE handle = INVALID_HANDLE_VALUE;
        ULONG length = 0;
        BYTE buffer[512];
        VIRTIO_PORT_INFO* port_info = NULL;

        SP_DEVICE_INTERFACE_DATA did_data;
        did_data.cbSize = sizeof(did_data);
        status = SetupDiEnumDeviceInterfaces(dev_info,
                                             0,
                                             &VIRTIO_SERIAL_GUID,
                                             index,
                                             &did_data);
        index++;
        if(status == FALSE)
        {
            /*
            If we hit this then we have reached the end of the
            list of serial devices.
            */
            fprintf(stderr, "%s not found\n", name);
            fprintf(stderr, "Turn UAC off and make sure that %s is defined in the VM config.\n", name);
            break;
        }

        status = SetupDiGetDeviceInterfaceDetail(dev_info,
                                                 &did_data,
                                                 NULL,
                                                 0,
                                                 &size,
                                                 NULL);

        detail_data = (SP_DEVICE_INTERFACE_DETAIL_DATA*)malloc(size);
        detail_data->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

        status = SetupDiGetDeviceInterfaceDetail(dev_info,
                                                 &did_data,
                                                 detail_data,
                                                 size,
                                                 &required_size,
                                                 NULL);
        if(status == FALSE)
        {
            DWORD error = GetLastError();
            fprintf(stderr, "SetupDiGetDeviceInterfaceDetail Error %d\n", error);
            free(detail_data);
            continue;
        }

        /* copy the device path */
        wcsncpy_s(current_device_path, MAX_PATH, detail_data->DevicePath, MAX_PATH - 1);
        free(detail_data);

        /* open the device to read the channel name */
        handle = CreateFile(current_device_path,
                            GENERIC_READ,
                            0,
                            NULL,
                            OPEN_EXISTING,
                            FILE_ATTRIBUTE_NORMAL,
                            NULL);
        if(handle == INVALID_HANDLE_VALUE)
        {
            fprintf(stderr, "CreateFile error %d\n", GetLastError());
            continue;
        }

        status = DeviceIoControl(handle,
                                 IOCTL_GET_INFORMATION,
                                 NULL,
                                 0,
                                 buffer,
                                 sizeof(buffer),
                                 &length,
                                 NULL);
        CloseHandle(handle);
        if(status == FALSE)
        {
            DWORD error = GetLastError();
            fprintf(stderr, "DeviceIoControl Error %d\n", error);
            continue;
        }

        port_info = (VIRTIO_PORT_INFO*)buffer;
        if(!strcmp(name, port_info->name))
        {
            if(wcslen(current_device_path) > (device_path_size + 1))
            {
                break;
            }
            
            /* if the channel name matches, copy out the device path and return success */
            wcscpy_s(device_path, device_path_size, current_device_path);
            success = TRUE;
            break;
        }
    }

    SetupDiDestroyDeviceInfoList(dev_info);
    return success;
}

int main(int argc, char* argv[])
{
    if(argc < 3 || (strcmp(argv[1], "r") != 0 && strcmp(argv[1], "w") != 0))
    {
        fprintf(stderr, "Usage: vmc-cat [r|w] [channel name]\n");
        return 1;
    }

    wchar_t device_path[MAX_PATH];
    BOOL success = virtioserial_find(argv[2], device_path, MAX_PATH);
    if(success == FALSE)
    {
        fprintf(stderr, "Unable to find %s\n", argv[2]);
        return 2;
    }

    DWORD mode = strcmp(argv[1], "r") == 0 ? GENERIC_READ : GENERIC_WRITE;
    HANDLE handle = CreateFile(device_path,
                               mode,
                               0,
                               NULL,
                               OPEN_EXISTING,
							   FILE_ATTRIBUTE_NORMAL | FILE_FLAG_NO_BUFFERING,
                               NULL);
	if(handle == INVALID_HANDLE_VALUE)
    {
        fprintf(stderr, "CreateFile error %d\n", GetLastError());
		return 3;
    }

    char buffer[4096];
    DWORD bytes_read = 0;
    DWORD bytes_written = 0;
    HANDLE in = mode == GENERIC_READ ? handle : GetStdHandle(STD_INPUT_HANDLE);
    HANDLE out = mode == GENERIC_WRITE ? handle : GetStdHandle(STD_OUTPUT_HANDLE);
    do
    {
        success = ReadFile(in, buffer, sizeof(buffer), &bytes_read, NULL);
        if(success == TRUE)
        {
            fprintf(stderr, "Executing WriteFile\n");
            success = WriteFile(out, buffer, bytes_read, &bytes_written, NULL);
            fprintf(stderr, "Finished WriteFile\n");
            if(success == FALSE)
            {
                fprintf(stderr, "WriteFile error %d\n", GetLastError());
                return 4;
            }
        }
    } while(success);
    
    DWORD error = GetLastError();
    if(error != ERROR_BROKEN_PIPE)
    {
        fprintf(stderr, "ReadFile error %d\n", error);
        return 5;
    }

    return 0;
}

